from typing import List, SupportsInt, Union

from .exceptions import InvalidOperationError
from .integer import IntBase, IntType, IntVar, uint8
from .utils import get_type


class VirtualPointer:
    """Provide pointer operation on bytearray.

    Args:
        source: The source ``bytearray`` to be read / write.
        data_type: The type of operated data. If it is ``str``, it will use
            ``utils.get_type`` to look up the type.
        offset: The distance from beginning to operating position.
    """

    def __init__(
        self,
        source: bytearray,
        data_type: Union[IntType, str] = uint8,
        offset: int = 0,
    ):
        self.source = source
        self.offset = offset
        self.data_type = data_type

    def __add__(self, other):
        """Support addition."""
        return self.add(other)

    def __sub__(self, other):
        """Support subtraction."""
        return self.sub(other)

    @property
    def data_type(self):
        """Get data type."""
        return self._data_type

    @data_type.setter
    def data_type(self, type_or_name: Union[IntType, str]):
        """Set data type."""
        if isinstance(type_or_name, str):
            try:
                self._data_type = get_type(type_name=type_or_name)

            except ValueError as e:
                raise InvalidOperationError("Unsupported type") from e

        elif isinstance(type_or_name, type) and issubclass(type_or_name, IntBase):
            self._data_type = type_or_name

        else:
            raise TypeError("Invalid type")

    def copy(self) -> "VirtualPointer":
        """Copy this object.

        The new object and the old object will operate on the same ``bytearray``.
        """
        return self.__class__(
            source=self.source,
            data_type=self.data_type,
            offset=self.offset,
        )

    def add(self, num: int) -> "VirtualPointer":
        """Offset this pointer position."""
        obj = self.copy()
        obj.offset += num * self.data_type.get_size()
        return obj

    def sub(self, num: int) -> "VirtualPointer":
        """Reverse offset this pointer position."""
        return self.add(-num)

    def cast(self, data_type: Union[IntType, str]) -> "VirtualPointer":
        """Cast to the specified type."""
        obj = self.copy()
        obj.data_type = data_type
        return obj

    def read_bytes(self, size: int) -> bytes:
        """Read bytes from source ``bytearray``."""
        if self.offset + size > len(self.source):
            raise InvalidOperationError("Read out of range")

        return bytes(self.source[self.offset : self.offset + size])

    def write_bytes(self, data: Union[bytes, bytearray, List[SupportsInt]]):
        """Write bytes into source ``bytearray``."""
        try:
            for i, v in enumerate(data):
                self.source[self.offset + i] = int(v)

        except IndexError as e:
            raise InvalidOperationError("Write out of range") from e

    def read(self) -> IntVar:
        """Read an integer from source ``bytearray``."""
        data = self.read_bytes(self.data_type.get_size())
        return self.data_type.from_bytes(data)

    def write(self, value: SupportsInt):
        """Write an integer into source ``bytearray``."""
        data = self.data_type(value).to_bytes()
        self.write_bytes(data)


def vptr(source: bytearray, data_type: Union[IntType, str] = uint8) -> VirtualPointer:
    """Shorthand of `VirtualPointer(source, data_type)`."""
    return VirtualPointer(source=source, data_type=data_type)



"""Implement functions which are used in the code decompiled by Ghidra."""

from typing import Type, TypeVar

from ..integer import IntVar, int8, int16, int32, int64, uint8, uint16, uint32, uint64
from ..utils import get_type
from .ida import _truncate, cfadd, ofsub, ofadd

# Refer to https://github.com/NationalSecurityAgency/ghidra/blob/master/Ghidra/Features/Decompiler/src/main/help/help/topics/DecompilePlugin/DecompilerConcepts.html    # noqa: E501

_T = TypeVar("_T", int8, int16, int32, int64, uint8, uint16, uint32, uint64)


def _zero_extend(x: IntVar, to_type: Type[_T]) -> _T:
    """Zero extend."""
    return to_type.from_bytes(x.to_bytes())


def _sign_extend(x: IntVar, to_type: Type[_T]) -> _T:
    """Sign extend."""
    t1 = get_type(size=x.get_size(), signed=True)
    t2 = get_type(size=to_type.get_size(), signed=True)
    return to_type.from_bytes(t2.from_bytes(t1(x).to_bytes()).to_bytes())


def sub21(x: uint16, c: int) -> uint8:
    """Implementation of `SUB21`."""
    return _truncate(x, c, uint8)


def sub41(x: uint32, c: int) -> uint8:
    """Implementation of `SUB41`."""
    return _truncate(x, c, uint8)


def sub42(x: uint32, c: int) -> uint16:
    """Implementation of `SUB42`."""
    return _truncate(x, c, uint16)


def sub81(x: uint64, c: int) -> uint8:
    """Implementation of `SUB81`."""
    return _truncate(x, c, uint8)


def sub82(x: uint64, c: int) -> uint16:
    """Implementation of `SUB82`."""
    return _truncate(x, c, uint16)


def sub84(x: uint64, c: int) -> uint32:
    """Implementation of `SUB84`."""
    return _truncate(x, c, uint32)


def zext12(x: uint8) -> uint16:
    """Implementation of `ZEXT12`."""
    return _zero_extend(x, uint16)


def zext14(x: uint8) -> uint32:
    """Implementation of `ZEXT14`."""
    return _zero_extend(x, uint32)


def zext18(x: uint8) -> uint64:
    """Implementation of `ZEXT18`."""
    return _zero_extend(x, uint64)


def zext24(x: uint16) -> uint32:
    """Implementation of `ZEXT24`."""
    return _zero_extend(x, uint32)


def zext28(x: uint16) -> uint64:
    """Implementation of `ZEXT28`."""
    return _zero_extend(x, uint64)


def zext48(x: uint32) -> uint64:
    """Implementation of `ZEXT48`."""
    return _zero_extend(x, uint64)


def sext12(x: uint8) -> uint16:
    """Implementation of `SEXT12`."""
    return _sign_extend(x, uint16)


def sext14(x: uint8) -> uint32:
    """Implementation of `SEXT14`."""
    return _sign_extend(x, uint32)


def sext18(x: uint8) -> uint64:
    """Implementation of `SEXT18`."""
    return _sign_extend(x, uint64)


def sext24(x: uint16) -> uint32:
    """Implementation of `SEXT24`."""
    return _sign_extend(x, uint32)


def sext28(x: uint16) -> uint64:
    """Implementation of `SEXT28`."""
    return _sign_extend(x, uint64)


def sext48(x: uint32) -> uint64:
    """Implementation of `SEXT48`."""
    return _sign_extend(x, uint64)


def sborrow1(x: uint8, y: uint8) -> int:
    """Implementation of `SBORROW1`."""
    return ofsub(x, y)


def sborrow2(x: uint16, y: uint16) -> int:
    """Implementation of `SBORROW2`."""
    return ofsub(x, y)


def sborrow4(x: uint32, y: uint32) -> int:
    """Implementation of `SBORROW4`."""
    return ofsub(x, y)


def sborrow8(x: uint64, y: uint64) -> int:
    """Implementation of `SBORROW8`."""
    return ofsub(x, y)


def carry1(x: uint8, y: uint8) -> int:
    """Implementation of `CARRY1`."""
    return cfadd(x, y)


def carry2(x: uint16, y: uint16) -> int:
    """Implementation of `CARRY2`."""
    return cfadd(x, y)


def carry4(x: uint32, y: uint32) -> int:
    """Implementation of `CARRY4`."""
    return cfadd(x, y)


def carry8(x: uint64, y: uint64) -> int:
    """Implementation of `CARRY8`."""
    return cfadd(x, y)


def scarry1(x: int8, y: int8) -> int:
    """Implementation of `SCARRY1`."""
    return ofadd(x, y)


def scarry2(x: int16, y: int16) -> int:
    """Implementation of `SCARRY2`."""
    return ofadd(x, y)


def scarry4(x: int32, y: int32) -> int:
    """Implementation of `SCARRY4`."""
    return ofadd(x, y)


def scarry8(x: int64, y: int64) -> int:
    """Implementation of `SCARRY8`."""
    return ofadd(x, y)



import ctypes
import re
import sys
from ctypes import (
    c_int8,
    c_int16,
    c_int32,
    c_int64,
    c_uint8,
    c_uint16,
    c_uint32,
    c_uint64,
)
from typing import ClassVar, Iterable, SupportsBytes, SupportsInt, Type, Union

from . import endian

if sys.version_info >= (3, 8):
    from typing import SupportsIndex
else:
    from typing_extensions import SupportsIndex

_CtypesInt = Union[
    c_int8, c_int16, c_int32, c_int64, c_uint8, c_uint16, c_uint32, c_uint64
]


class _UnaryOp:
    def __call__(self) -> "IntVar":
        ...


class _BinaryOp:
    def __call__(self, other: SupportsInt) -> "IntVar":
        ...


class _ComparisonOp:
    def __call__(self, other: SupportsInt) -> bool:
        ...


# Operation and comparison methods of integer types.

_INT_OP = [
    "__neg__",
    "__pos__",
    "__abs__",
    "__add__",
    "__sub__",
    "__mul__",
    "__truediv__",
    "__floordiv__",
    "__mod__",
    "__invert__",
    "__and__",
    "__or__",
    "__xor__",
    "__lshift__",
    "__rshift__",
]

_INT_CMP = ["__gt__", "__ge__", "__eq__", "__le__", "__lt__", "__ne__"]


class IntMeta(type):
    """Metaclass of integer type."""

    _base: Type[_CtypesInt]
    _size: ClassVar[int]
    _signed: ClassVar[bool]
    _impl: _CtypesInt

    def __init__(cls, name, bases, attr_dict):
        super().__init__(name, bases, attr_dict)

        base = getattr(ctypes, "c_%s" % cls.__name__.lower())
        setattr(cls, "_base", base)

        size = int(re.match(r"(U*)Int(\d+)", cls.__name__).group(2)) // 8
        setattr(cls, "_size", size)

        signed = bool(re.match(r"Int\d+", cls.__name__))
        setattr(cls, "_signed", signed)

        for op in _INT_OP:
            setattr(cls, op, cls._build_operation(op))

        for cmp in _INT_CMP:
            setattr(cls, cmp, cls._build_comparison(cmp))

    @staticmethod
    def _build_operation(func_name):
        """Build operation method.

        If it is a unary operation, the result is still the own type.
        If the type of another operand is ``Integer`` and is inconsistent with
        the own type, they result will be converted to larger size or unsigned type.
        If another operand is ``int`` or can be converted to ``int``, the result
        will be converted to the own type.
        """

        def decorator(self, *args):
            data_type = type(self)

            try:
                operator = getattr(int, func_name)
            except AttributeError:
                return NotImplemented

            # If a binary operation
            if args:
                other = args[0]

                if isinstance(other, IntBase):
                    # If their sizes are equal, the type of result is unsigned.
                    if self.get_size() == other.get_size():
                        data_type = type(other if self.get_signed() else self)

                    # If their sizes are not equal, the type of result is
                    # larger size type.
                    else:
                        if self.get_size() < other.get_size():
                            data_type = type(other)
                            self, other = other, self

                        else:
                            data_type = type(self)

                elif not hasattr(other, "__int__"):
                    return NotImplemented

                return data_type(operator(self._impl.value, int(other)))

            return data_type(operator(self._impl.value))

        return decorator

    @staticmethod
    def _build_comparison(func_name):
        """Build comparison method."""

        def decorator(self, other):
            func = getattr(int, func_name)
            return func(self._impl.value, int(other))

        return decorator


class _IntBase:
    """Type hints of integer type."""

    _base: Type[_CtypesInt]
    _size: ClassVar[int]
    _signed: ClassVar[bool]
    _impl: _CtypesInt

    __neg__: _UnaryOp
    __pos__: _UnaryOp
    __abs__: _UnaryOp
    __add__: _BinaryOp
    __radd__: _BinaryOp
    __sub__: _BinaryOp
    __rsub__: _BinaryOp
    __mul__: _BinaryOp
    __rmul__: _BinaryOp
    __truediv__: _BinaryOp
    __rtruediv__: _BinaryOp
    __floordiv__: _BinaryOp
    __rfloordiv__: _BinaryOp
    __mod__: _BinaryOp
    __rmod__: _BinaryOp
    __invert__: _UnaryOp
    __and__: _BinaryOp
    __rand__: _BinaryOp
    __or__: _BinaryOp
    __ror__: _BinaryOp
    __xor__: _BinaryOp
    __rxor__: _BinaryOp
    __lshift__: _BinaryOp
    __rlshift__: _BinaryOp
    __rshift__: _BinaryOp
    __rrshift__: _BinaryOp
    __gt__: _ComparisonOp
    __ge__: _ComparisonOp
    __le__: _ComparisonOp
    __lt__: _ComparisonOp


class IntBase(_IntBase):
    """Base class of integer type."""

    def __init__(self, x: SupportsInt):
        self._impl = self._base(int(x))

    def __int__(self) -> int:
        return int(self._impl.value)

    @classmethod
    def get_size(cls) -> int:
        """Get size (bytes) of this type."""
        return cls._size

    @classmethod
    def get_signed(cls) -> bool:
        """Get signed of this type."""
        return cls._signed

    @classmethod
    def from_bytes(cls, data: Union[Iterable[SupportsIndex], SupportsBytes]):
        """Return a value of this type from given bytes"""
        return cls(
            int.from_bytes(data, byteorder=endian.BYTE_ORDER, signed=cls.get_signed())
        )

    def to_bytes(self) -> bytes:
        """Covert this value to bytes."""
        return int(self).to_bytes(
            length=self.get_size(),
            byteorder=endian.BYTE_ORDER,
            signed=self.get_signed(),
        )


class Int8(IntBase, metaclass=IntMeta):
    """Int8"""


class Int16(IntBase, metaclass=IntMeta):
    """Int16"""


class Int32(IntBase, metaclass=IntMeta):
    """Int32"""


class Int64(IntBase, metaclass=IntMeta):
    """Int64"""


class UInt8(IntBase, metaclass=IntMeta):
    """UInt8"""


class UInt16(IntBase, metaclass=IntMeta):
    """UInt16"""


class UInt32(IntBase, metaclass=IntMeta):
    """UInt32"""


class UInt64(IntBase, metaclass=IntMeta):
    """UInt64"""


int8 = Int8
int16 = Int16
int32 = Int32
int64 = Int64
uint8 = UInt8
uint16 = UInt16
uint32 = UInt32
uint64 = UInt64

# Used as typing
IntVar = IntBase
IntType = Type[IntVar]



import sys

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

BIG_ENDIAN: Literal["big"] = "big"
LITTLE_ENDIAN: Literal["little"] = "little"

# Default use little endian.
BYTE_ORDER: Literal["big", "little"] = LITTLE_ENDIAN



"""Implement functions which are used in the code decompiled by IDA."""

from typing import Type, TypeVar

from .. import endian
from ..integer import (
    IntType,
    IntVar,
    int8,
    int16,
    int32,
    int64,
    uint8,
    uint16,
    uint32,
    uint64,
)
from ..utils import get_type

_T = TypeVar("_T", int8, int16, int32, int64, uint8, uint16, uint32, uint64)

# Refer to defs.h of IDA.


def _truncate(x: IntVar, c: int, to_type: Type[_T]) -> _T:
    """Truncate data."""
    data = x.to_bytes()
    to_size = to_type.get_size()
    return to_type.from_bytes(data[c : c + to_size])


def last_ind(x: IntVar, part_type: IntType) -> int:
    """Implementation of `LAST_IND`."""
    return x.get_size() // part_type.get_size() - 1


def low_ind(x: IntVar, part_type: IntType) -> int:
    """Implementation of `LOW_IND`."""
    return last_ind(x, part_type) if endian.BYTE_ORDER == endian.BIG_ENDIAN else 0


def high_ind(x: IntVar, part_type: IntType) -> int:
    """Implementation of `HIGH_IND`."""
    return 0 if endian.BYTE_ORDER == endian.BIG_ENDIAN else last_ind(x, part_type)


def byten(x: IntVar, n: int) -> uint8:
    """Implementation of `BYTEn`."""
    return _truncate(x, n * 1, uint8)


def wordn(x: IntVar, n: int) -> uint16:
    """Implementation of `WORDn`."""
    return _truncate(x, n * 2, uint16)


def dwordn(x: IntVar, n: int) -> uint32:
    """Implementation of `DWORDn`."""
    return _truncate(x, n * 4, uint32)


def lobyte(x: IntVar) -> uint8:
    """Implementation of `LOBYTE`."""
    return byten(x, low_ind(x, uint8))


def loword(x: IntVar) -> uint16:
    """Implementation of `LOWORD`."""
    return wordn(x, low_ind(x, uint16))


def lodword(x: IntVar) -> uint32:
    """Implementation of `LODWORD`."""
    return dwordn(x, low_ind(x, uint32))


def hibyte(x: IntVar) -> uint8:
    """Implementation of `HIBYTE`."""
    return byten(x, high_ind(x, uint8))


def hiword(x: IntVar) -> uint16:
    """Implementation of `HIWORD`."""
    return wordn(x, high_ind(x, uint16))


def hidword(x: IntVar) -> uint32:
    """Implementation of `HIDWORD`."""
    return dwordn(x, high_ind(x, uint32))


def byte1(x: IntVar) -> uint8:
    """Implementation of `BYTE1`."""
    return byten(x, 1)


def byte2(x: IntVar) -> uint8:
    """Implementation of `BYTE2`."""
    return byten(x, 2)


def byte3(x: IntVar) -> uint8:
    """Implementation of `BYTE3`."""
    return byten(x, 3)


def byte4(x: IntVar) -> uint8:
    """Implementation of `BYTE4`."""
    return byten(x, 4)


def byte5(x: IntVar) -> uint8:
    """Implementation of `BYTE5`."""
    return byten(x, 5)


def byte6(x: IntVar) -> uint8:
    """Implementation of `BYTE6`."""
    return byten(x, 6)


def byte7(x: IntVar) -> uint8:
    """Implementation of `BYTE7`."""
    return byten(x, 7)


def byte8(x: IntVar) -> uint8:
    """Implementation of `BYTE8`."""
    return byten(x, 8)


def byte9(x: IntVar) -> uint8:
    """Implementation of `BYTE9`."""
    return byten(x, 9)


def byte10(x: IntVar) -> uint8:
    """Implementation of `BYTE10`."""
    return byten(x, 10)


def byte11(x: IntVar) -> uint8:
    """Implementation of `BYTE11`."""
    return byten(x, 11)


def byte12(x: IntVar) -> uint8:
    """Implementation of `BYTE12`."""
    return byten(x, 12)


def byte13(x: IntVar) -> uint8:
    """Implementation of `BYTE13`."""
    return byten(x, 13)


def byte14(x: IntVar) -> uint8:
    """Implementation of `BYTE14`."""
    return byten(x, 14)


def byte15(x: IntVar) -> uint8:
    """Implementation of `BYTE15`."""
    return byten(x, 15)


def word1(x: IntVar) -> uint16:
    """Implementation of `WORD1`."""
    return wordn(x, 1)


def word2(x: IntVar) -> uint16:
    """Implementation of `WORD2`."""
    return wordn(x, 2)


def word3(x: IntVar) -> uint16:
    """Implementation of `WORD3`."""
    return wordn(x, 3)


def word4(x: IntVar) -> uint16:
    """Implementation of `WORD4`."""
    return wordn(x, 4)


def word5(x: IntVar) -> uint16:
    """Implementation of `WORD5`."""
    return wordn(x, 5)


def word6(x: IntVar) -> uint16:
    """Implementation of `WORD6`."""
    return wordn(x, 6)


def word7(x: IntVar) -> uint16:
    """Implementation of `WORD7`."""
    return wordn(x, 7)


def dword1(x: IntVar) -> uint32:
    """Implementation of `DWORD1`."""
    return dwordn(x, 1)


def dword2(x: IntVar) -> uint32:
    """Implementation of `DWORD2`."""
    return dwordn(x, 2)


def dword3(x: IntVar) -> uint32:
    """Implementation of `DWORD3`."""
    return dwordn(x, 3)


def sbyten(x: IntVar, n: int) -> int8:
    """Implementation of `SBYTEn`."""
    return _truncate(x, n * 1, int8)


def swordn(x: IntVar, n: int) -> int16:
    """Implementation of `SWORDn`."""
    return _truncate(x, n * 2, int16)


def sdwordn(x: IntVar, n: int) -> int32:
    """Implementation of `SDWORDn`."""
    return _truncate(x, n * 4, int32)


def slobyte(x: IntVar) -> int8:
    """Implementation of `SLOBYTE`."""
    return sbyten(x, low_ind(x, int8))


def sloword(x: IntVar) -> int16:
    """Implementation of `SLOWORD`."""
    return swordn(x, low_ind(x, int16))


def slodword(x: IntVar) -> int32:
    """Implementation of `SLODWORD`."""
    return sdwordn(x, low_ind(x, int32))


def shibyte(x: IntVar) -> int8:
    """Implementation of `SHIBYTE`."""
    return sbyten(x, high_ind(x, int8))


def shiword(x: IntVar) -> int16:
    """Implementation of `SHIWORD`."""
    return swordn(x, high_ind(x, int16))


def shidword(x: IntVar) -> int32:
    """Implementation of `SHIDWORD`."""
    return sdwordn(x, high_ind(x, int32))


def sbyte1(x: IntVar) -> int8:
    """Implementation of `SBYTE1`."""
    return sbyten(x, 1)


def sbyte2(x: IntVar) -> int8:
    """Implementation of `SBYTE2`."""
    return sbyten(x, 2)


def sbyte3(x: IntVar) -> int8:
    """Implementation of `SBYTE3`."""
    return sbyten(x, 3)


def sbyte4(x: IntVar) -> int8:
    """Implementation of `SBYTE4`."""
    return sbyten(x, 4)


def sbyte5(x: IntVar) -> int8:
    """Implementation of `SBYTE5`."""
    return sbyten(x, 5)


def sbyte6(x: IntVar) -> int8:
    """Implementation of `SBYTE6`."""
    return sbyten(x, 6)


def sbyte7(x: IntVar) -> int8:
    """Implementation of `SBYTE7`."""
    return sbyten(x, 7)


def sbyte8(x: IntVar) -> int8:
    """Implementation of `SBYTE8`."""
    return sbyten(x, 8)


def sbyte9(x: IntVar) -> int8:
    """Implementation of `SBYTE9`."""
    return sbyten(x, 9)


def sbyte10(x: IntVar) -> int8:
    """Implementation of `SBYTE10`."""
    return sbyten(x, 10)


def sbyte11(x: IntVar) -> int8:
    """Implementation of `SBYTE11`."""
    return sbyten(x, 11)


def sbyte12(x: IntVar) -> int8:
    """Implementation of `SBYTE12`."""
    return sbyten(x, 12)


def sbyte13(x: IntVar) -> int8:
    """Implementation of `SBYTE13`."""
    return sbyten(x, 13)


def sbyte14(x: IntVar) -> int8:
    """Implementation of `SBYTE14`."""
    return sbyten(x, 14)


def sbyte15(x: IntVar) -> int8:
    """Implementation of `SBYTE15`."""
    return sbyten(x, 15)


def sword1(x: IntVar) -> int16:
    """Implementation of `SWORD1`."""
    return swordn(x, 1)


def sword2(x: IntVar) -> int16:
    """Implementation of `SWORD2`."""
    return swordn(x, 2)


def sword3(x: IntVar) -> int16:
    """Implementation of `SWORD3`."""
    return swordn(x, 3)


def sword4(x: IntVar) -> int16:
    """Implementation of `SWORD4`."""
    return swordn(x, 4)


def sword5(x: IntVar) -> int16:
    """Implementation of `SWORD5`."""
    return swordn(x, 5)


def sword6(x: IntVar) -> int16:
    """Implementation of `SWORD6`."""
    return swordn(x, 6)


def sword7(x: IntVar) -> int16:
    """Implementation of `SWORD7`."""
    return swordn(x, 7)


def sdword1(x: IntVar) -> int32:
    """Implementation of `SDWORD1`."""
    return sdwordn(x, 1)


def sdword2(x: IntVar) -> int32:
    """Implementation of `SDWORD2`."""
    return sdwordn(x, 2)


def sdword3(x: IntVar) -> int32:
    """Implementation of `SDWORD3`."""
    return sdwordn(x, 3)


def pair(high: IntVar, low: IntVar) -> IntVar:
    """Implementation of `__PAIR__`."""
    size = high.get_size()
    signed = high.get_signed()
    int_type = get_type(size=size * 2, signed=signed)
    return int_type(high) << size * 8 | type(high)(low)


def rol(value: IntVar, count: int) -> IntVar:
    """Implementation of `__ROL__`."""
    data_type = type(value)
    nbits = value.get_size() * 8

    if count > 0:
        count %= nbits
        high = value >> (nbits - count)
        if value.get_signed():
            high &= ~(data_type(-1) << count)
        value <<= count
        value |= high

    else:
        count = -count % nbits
        low = value << (nbits - count)
        value >>= count
        value |= low

    return value


def rol1(value: uint8, count: int) -> uint8:
    """Implementation of `__ROL1__`."""
    return uint8(rol(value, count))


def rol2(value: uint16, count: int) -> uint16:
    """Implementation of `__ROL2__`."""
    return uint16(rol(value, count))


def rol4(value: uint32, count: int) -> uint32:
    """Implementation of `__ROL4__`."""
    return uint32(rol(value, count))


def rol8(value: uint64, count: int) -> uint64:
    """Implementation of `__ROL8__`."""
    return uint64(rol(value, count))


def ror1(value: uint8, count: int) -> uint8:
    """Implementation of `__ROR1__`."""
    return uint8(rol(value, -count))


def ror2(value: uint16, count: int) -> uint16:
    """Implementation of `__ROR2__`."""
    return uint16(rol(value, -count))


def ror4(value: uint32, count: int) -> uint32:
    """Implementation of `__ROR4__`."""
    return uint32(rol(value, -count))


def ror8(value: uint64, count: int) -> uint64:
    """Implementation of `__ROR8__`."""
    return uint64(rol(value, -count))


def mkcshl(value: IntVar, count: int) -> int:
    """Implementation of `__MKCSHL__`."""
    nbits = value.get_size() * 8
    count %= nbits
    return int((value >> (nbits - count)) & 1)


def mkcshr(value: IntVar, count: int) -> int:
    """Implementation of `__MKCSHR__`."""
    return int((value >> (count - 1)) & 1)


def sets(x: IntVar) -> int:
    """Implementation of `__SETS__`."""
    data_type = get_type(size=x.get_size(), signed=True)
    return int(data_type(x) < 0)


def ofsub(x: IntVar, y: IntVar) -> int:
    """Implementation of `__OFSUB__`."""
    if x.get_size() < y.get_size():
        x2 = x
        sx = sets(x2)
        return int((sx ^ sets(y)) & (sx ^ sets(x2 - y)))
    else:
        y2 = y
        sx = sets(x)
        return int((sx ^ sets(y2)) & (sx ^ sets(x - y2)))


def ofadd(x: IntVar, y: IntVar) -> int:
    """Implementation of `__OFADD__`."""
    if x.get_size() < y.get_size():
        x2 = x
        sx = sets(x2)
        return int(((1 ^ sx) ^ sets(y)) & (sx ^ sets(x2 + y)))
    else:
        y2 = y
        sx = sets(x)
        return int(((1 ^ sx) ^ sets(y2)) & (sx ^ sets(x + y2)))


def cfsub(x: IntVar, y: IntVar) -> int:
    """Implementation of `__CFSUB__`."""
    size = max(x.get_size(), y.get_size())
    data_type = get_type(size=size, signed=False)
    return int(data_type(x) < data_type(y))


def cfadd(x: IntVar, y: IntVar) -> int:
    """Implementation of `__CFADD__`."""
    size = max(x.get_size(), y.get_size())
    data_type = get_type(size=size, signed=False)
    return int(data_type(x) > data_type(x + y))


# Refer to https://gcc.gnu.org/onlinedocs/gcc/Other-Builtins.html.


def bswap32(value: uint32) -> uint32:
    """Implementation of `bswap32`."""
    return value.from_bytes(value.to_bytes()[::-1])


def clz(x: IntVar) -> int:
    """Implementation of `__clz`."""
    return x.get_size() * 8 - len(bin(int(x))[2:])



import re
from typing import Optional

from .integer import IntType, int8, int16, int32, int64, uint8, uint16, uint32, uint64


def get_type(
    size: Optional[int] = None,
    signed: Optional[bool] = None,
    type_name: Optional[str] = None,
) -> IntType:
    """Get type int with specified size and signed.

    Args:
        size: The size of the type.
        signed: Is the type signed.
        type_name: The lowercase name of type to find. If it is None,
            ``size`` and ``signed`` must be given.

    Raises:
        ValueError: If no matched type.
    """
    int_types = [int8, int16, int32, int64, uint8, uint16, uint32, uint64]

    if type_name is not None:
        match = re.match(r"(u*)int(\d+)", type_name)

        if match:
            nbits = int(match.group(2))
            if nbits % 8 == 0:
                signed = not bool(match.group(1))
                size = nbits // 8

    if size is not None and signed is not None:
        for int_type in int_types:
            if int_type.get_size() == size and int_type.get_signed() == signed:
                return int_type

    raise ValueError("No matched type")
